#+++
#  title:	Mkctable.awk
#  abstract:	build character tables for JOVE from a character description
#  author:	T.R.Hageman, Groningen, The Netherlands.
#  created:	march 1992
#  modified:
#  description:
#  	usage: awk -f Mkctable.awk [options] <chardesc> <tabledesc>...
#
#	Options:
#
#	  start=<num>	- set lower bound of generated tables to <num>
#			  (between 0 and 255, default 0)
#	  end=<num>	- set upper bound of generated tables to <num>
#			  (between 1 and 256, default 256)
#	  out=<file>	- set output filename to <file>
#			  (default = name of last <tabledesc> file, with
#			   extension replaced by ".tab")
#	  usebounds=1	- take bounds from directives in <tabledesc> file
#	  verbose=1	- show progress
#
#	The file <chardesc> contains the descriptions of the characters.
#	A character description is a line with the following format:
#
#	<name> <class> <word> [<lower> [<equiv> [<accent> [<aliases>]]]]
#
#	where:
#
#	<name> 	the name by which the character is known.
#		This can be a string of arbitrary length (although for
#		the sake of alignment names shorter that 7 characters are
#		preferred) consisting of arbitrary characters except
#		whitespace.
#		Caveat:
#		  [single characters]
#			- these are treated specially in the output of
#			  the tables: instead of the octal character
#			  code the character itself is printed.
#		  [names starting with `=']
#			- are treated as if it were a single character,
#			  that is, the remainder of the name is printed
#			  literally instead of the octal character code.
#			  This can be used for characters that need to
#			  be escaped in C character constants (like [\]
#		          and [']), or for characters that have special
#			  meaning in this script (like [=] and [#])
#
#	<class>	the character's class, in terms of:
#		  _U  - upper case alphabetic
#		  _L  - lower case alphabetic
#		  _N  - numeral digit
#		  _S  - whitespace
#		  _C  - control character
#		  _P  - punctuation
#		  _Op - Open bracket
#		  _Cl - Close bracket
#		separated with the `|' character.
#
#	<word>	the modes in which the character is considered
#		part of a word, in terms of:
#		  F   - Fundamental mode
#		  T   - Text mode
#		  C   - C mode
#		  L   - Lisp mode
#		  P   - Perl mode
#		separated with the `|' character, or one of
#		  0   - never a word in any mode
#		  ALL - a word in all modes.
#
#	<lower>	the name of the lowercase character (that is the character
#		which should be the result of tolower(thischar)),
#		or `=' if it is the same as the character itself.
#
#	<equiv>	the name of the equivalence character (that is the character
#		which it should match in a case-insensitive search),
#		or `=' if it is the same as the character itself.
#
#	<accent> the accent sequence to produce this character (that is a
#		character pair (base,accented) to produce ``exotic''
#		characters.)  Note that `base' is escaped if necessary,
#		but `accented' should be valid in a C string.
#
#	<aliases> a list of character names, separated with semicolon (;)
#		which are aliases for this character.  That is, if the
#		character is NOT defined in the character set, and a
#		character in its <aliases> list is defined, the	alias's
#		character code is entered in the accent table at the
#		character's accent sequence. (if multiple alias characters
#		were defined in the character set, the first match is taken)
#		An alias of the form "cond?alias" is added conditionally;
#		the <cond> character should exist.
#
#	The <tabledesc> files describe a specific character set.
#	It is just a sequence of character names, in the desired order.
#	Instead of a character name you can use `=' which means keep the
#	original character definition (character tables can be overlaid).
#
#	You can also specify the following directive lines in a <tabledesc>
#	file:
#
#	@<num>		- set the character code index to <num>,
#			  where <num> should be a number between 0 and 256.
#	@Name <name>	- set the character set's name to <name>.  It may
#			  contain spaces; double quotes should be escaped
#			  with a backslash (like \").
#	@ClassDefault [<class>]
#			- set default value for unused entries in CharTable
#			  (see the description of <class> above)
#	@WordDefault [<word>]
#			- set default value for unused entries in WordTable
#			  (see the description of <word> above)
#	{<start>	- set lower bound for partial output of tables.
#	}<end>		- set upper bound for partial output of tables.
#			  (These are usually ignored, except when
#			   "usebounds=1" is given on the command line.)
#
#	The output of the program are the contents of the following tables
#	in C-syntax:
#
#	  CharTable	- the character class table.
#	  WordTable	- the word class table.
#	  LocTable	- the lowercase conversion table.
#	  UpcTable	- the uppercase conversion table.
#	  CaseEquiv	- the equivalence table for case-independent search.
#	  AccentTab	- the accent table.
#	and
#	  charset_name	- the name of the character set.
#
#	The start- and end indices of the tables (except `AccentTab') can
#	optionally be specified on the command line as "start=M" and "end=N",
#	respectively.  Alternatively, if you specify "usebounds=1", the
#	bounds as specified in <tabledesc> -if any- are used, see above.
#
#	In both the <chardesc> and <tabledesc> files, lines starting with `#'
#	are considered comment lines and are ignored; empty lines are also
#	ignored.
#
#  bugs:
#	Open/close paren definitions should come from chartabs too.
#	(these are currently hardwired, 7-bit ASCII, in ../c.c)
#---

BEGIN {
	split("$Id: Mkctable.awk,v 1.14 1993/11/05 22:22:41 tom Exp tom $", Id)
	if (i = index(Id[2], ","))
		ScriptName = substr(Id[2], 1, i - 1)
	else if ((ScriptName = Id[2]) == "")
		ScriptName = "mkctable.awk"
	end = 256
}

Filename != FILENAME {
	Filename = FILENAME
	FileStart = NR - 1
	NFiles++;
	if (verbose != "") print FILENAME ":"
	Files = Files " " FILENAME
	ClassDefault = "_P"
	WordDefault = "0"
}

/^[ \t]*$/ { next }
/^#/ { next }
/^\$Id/ { if (NF >= 3) Files = Files "(" $3 ")"; next }

NFiles == 1 {
	# Note this is the first file so line numbers are correct
	if (CharDef == "") {
		if ($1 == "@CharDef") {
			CharDef = 1
			next
		}
		else {
			print FILENAME ": not a character definition file", \
				"(\"@CharDef\" missing)"
			exit
		}
	}
	if (NF < 3) {
		printf("%s:%d: too few fields\n", FILENAME, NR)
		fatal = 1
		next
	}
	if (NameDefinedAtLine[$1] != "") {
		printf("%s:%d: multiple definition of character `%s'\n", \
		       FILENAME, NR, $1)
		printf("%s:%d: (previous definition of this character)\n", \
		       FILENAME, NameDefinedAtLine[$1])
	}
	NameDefinedAtLine[$1] = NR

	name[Nnames++] = $1
	class[$1] = $2
	word[$1] = $3
	if (NF < 4 || $4 == "=") lower[$1] = $1; else lower[$1] = $4
	if (NF < 5 || $5 == "=") equiv[$1] = $1; else equiv[$1] = $5
	if (NF < 6) next
	if (AccentDefinedAtLine[$6] != "") {
		printf("%s:%d: multiple definition of accent `%s'\n", \
		       FILENAME, NR, $6)
		printf("%s:%d: (previous definition of this accent)\n", \
		       FILENAME, AccentDefinedAtLine[$6])
	}
	AccentDefinedAtLine[$6] = NR
	accent[$1] = $6
	if (length(accent[$1]) == 1) accent[$1] = accent[$1] " "
	if (NF < 7 || $7 ~ /^#.*/) next
	aliases[$1] = $7
	next
}

/^{[0-9]+$/ { if (usebounds != "") start = substr($1, 2) + 0; next }
/^}[0-9]+$/ { if (usebounds != "") end = substr($1, 2) + 0; next }

/^@[0-9]+$/ {
	cc = substr($0, 2) + 0
	next
}
/^@ClassDefault/ {
	if (NF < 2) ClassDefault = "_P"; else ClassDefault = $2
	next
}
/^@WordDefault/ {
	if (NF < 2) WordDefault = "0"; else WordDefault = $2
	next
}
/^@Name/ {
	ShortName = Name = $2
	for (i = 3; i <= NF; i++) Name = Name " " $i
	next
}
/^@/ {
	printf("%s:%d: unrecognized directive `%s'\n", \
	       FILENAME, NR - FileStart, $1)
	next
}

{
	for (i = 1;  i <= NF;  i++) {
		classdef[cc] = ClassDefault
		worddef[cc] = WordDefault
		if ($i == "" || $i == "=") {
			cc++
			continue
		}
		if (NameDefinedAtLine[$i] == "") {
			printf("%s:%d: character `%s' undefined\n", \
			       FILENAME, NR - FileStart, $i)
		}
		if (code[$i] != "" && cc != code[$i]) {
			printf("%s:%d: multiple use of character `%s'", \
			       FILENAME, NR - FileStart, $i)
			printf(" (at 0%o, was 0%o)\n", cc, code[$i])
		}
		code[$i] = cc;
		char[cc] = $i
		cc++
	}
}

cc > 256 {
	printf("%s:%d: too many entries\n", FILENAME, NR - FileStart)
}

END {
	if (out == "") {
		if (i = index(out = Filename, "."))
			out = substr(out, 1, i - 1)
		out = out ".tab"
	}
	print "[" Files " -> " out " ]"

	print "/* This file is generated using the awk script \"" \
	      ScriptName "\" (" Id[3] " " Id[4] ")" >out
	print "   from:" Files "\n */" >out

	printf("\nprivate const char charset_name[] = \"%s\";\n", Name) >out

	# construct uppercase table.  Invariant: ``upper[lower[c]] == c''
	for (i in lower)
		upper[i] = i;
	for (i in lower)
		if (lower[i] != i)
			upper[lower[i]] = i;

	if (debug != "") {
	    if (debug > 2) {
		print "DEBUG dump of character descriptions"
		print "NAME	CLASS	WORD	LOWER	UPPER	EQUIV	ACCENT"
		for (i = 0;  i < Nnames;  i++)
			print name[i] "\t" class[name[i]] "\t" word[name[i]] \
				"\t" lower[name[i]] "\t" upper[name[i]] \
				"\t" equiv[name[i]] "\t" accent[name[i]]
	    }
	    if (debug > 1) {
		print
		print "DEBUG dump of character set table"
		for (i = 0;  i < 256;  ) {
			printf("\t%s", char[i])
			if (code[char[i]] != i && code[char[i]] != 0)
				printf("(%03o)", code[char[i]])
			if (++i % 8 == 0) {
				print
				if (i % 64 == 0) print
			}
		}
		print
		print "DEBUG dump of class default table"
		for (i = 0;  i < 256;  ) {
			printf("\t%s", classdef[i])
			if (++i % 8 == 0) {
				print
				if (i % 64 == 0) print
			}
		}
		print
		print "DEBUG dump of word default table"
		for (i = 0;  i < 256;  ) {
			printf("\t%s", worddef[i])
			if (++i % 8 == 0) {
				print
				if (i % 64 == 0) print
			}
		}
		print
	    }
	}

	if8 = "#if (BPC > 7)"
	fi8 = "#endif /* (BPC > 7) */"

	openformat = "\npublic Const char %s[CTABLESIZE] = {" \
		     sprintf("	/* [ %d - %d ) */\n%s", \
			     start, end, \
			     substr("\t\t\t\t\t\t\t", 8 - start % 8))
	closeformat = "};\n"
	if (end > 128) 	closeformat = fi8 "\n" closeformat
	if (end % 8)	closeformat = "\n" closeformat

	printf(openformat, "CharTable") >out
	for (i = start;  i < end; ) {
		if ((c = class[char[i+0]]) == "" && (c = classdef[i+0]) == "") {
			if (debug == 0.5)
				print "ClassDefault! i =", i, \
					" char[i] =", char[i], \
					" class[char[i]] =", class[char[i]], \
					" classdef[i] =", classdef[i]
			c = ClassDefault
		}
		printf("\t%s,", c) >out
		if (++i % 8 == 0) {
			print >out
			if (i % 64 == 0 && i != end)
				if (i == 128) print if8 >out; else print >out
		}
	}
	printf(closeformat) >out

	printf(openformat, "WordTable") >out
	for (i = start;  i < end; ) {
		if ((c = word[char[i+0]]) == "" && (c = worddef[i+0]) == "")
			c = WordDefault
		printf("\t%s,", c) >out
		if (++i % 8 == 0) {
			print >out
			if (i % 64 == 0 && i != end)
				if (i == 128) print if8 >out; else print >out
		}
	}
	printf(closeformat) >out

	printf(openformat, "LocTable") >out
	for (i = start;  i < end; ) {
		if ((c = code[lower[char[i+0]]]) == "")
			c = i;
		if (length(char[c]) == 1)
			printf("\t'%s',", char[c]) >out
		else if (char[c] ~ /^=.*/)
			printf("\t'%s',", substr(char[c], 2)) >out
		else
			printf("\t'\\%03o',", c) >out
		if (++i % 8 == 0) {
			print >out
			if (i % 64 == 0 && i != end)
				if (i == 128) print if8 >out; else print >out
		}
	}
	printf(closeformat) >out

	printf(openformat, "UpcTable") >out
	for (i = start;  i < end; ) {
		if ((c = code[upper[char[i+0]]]) == "")
			c = i;
		if (length(char[c]) == 1)
			printf("\t'%s',", char[c]) >out
		else if (char[c] ~ /^=.*/)
			printf("\t'%s',", substr(char[c], 2)) >out
		else
			printf("\t'\\%03o',", c) >out
		if (++i % 8 == 0) {
			print >out
			if (i % 64 == 0 && i != end)
				if (i == 128) print if8 >out; else print >out
		}
	}
	printf(closeformat) >out
	printf("\n%s", if8) >out
	printf(openformat, "CaseEquiv") >out
	for (i = start;  i < end; ) {
		if ((c = code[equiv[char[i+0]]]) == "")
			c = i;
		if (length(char[c]) == 1)
			printf("\t'%s',", char[c]) >out
		else if (char[c] ~ /^=.*/)
			printf("\t'%s',", substr(char[c], 2)) >out
		else
			printf("\t'\\%03o',", c) >out
		if (++i % 8 == 0) {
			print >out
			if (i % 64 == 0 && i != end)
				if (i == 128) print if8 >out; else print >out
		}
	}
	printf(closeformat) >out

	# Now build the accent table
	for (i = 0;  i < Nnames;  i++) {

		if (accent[name[i]] == "")
			continue

		if ((charcode = code[name[i]]) == "") {

			if (aliases[name[i]] == "")
				continue

			# Handle ``aliased accents''

			if (debug > 0) printf("(aliases[%s] = %s)\n", name[i], aliases[name[i]])

			Nalts = split(aliases[name[i]], alts, ";")
			for (cc = 1;  cc <= Nalts;  cc++) {
				alt = alts[cc];
				if (debug > 0) printf(" (%s ", alt)
				if (c = index(alt, "?")) {	# conditional
					if (code[substr(alt, 1, c - 1)] == "") {
						if (debug > 0) printf("code[%s]=FALSE)\n", substr(alt, 1, c - 1))
						continue
					}
					if (debug > 0) printf("code[%s]=%d ", substr(alt, 1, c - 1), code[substr(alt, 1, c - 1)])
					alt = substr(alt, c + 1)
				}
				if (debug > 0) printf("code=%s accent=%s)\n", code[alt], accent[alt])
				if ((charcode = code[alt]) != "")
					break
			}
			if (charcode == "")
				continue
		}
		if (AccentTab[c = substr(accent[name[i]], 1, 1)] == "") {
			Accent[Naccents++] = c
			if (index("\"\\", c))
				AccentTab[c] = "\\" c
			else
				AccentTab[c] = c
		}
		AccentTab[c] = sprintf("%s%s\\%03o", AccentTab[c], \
				       substr(accent[name[i]], 2), \
				       charcode)
	}

	print "\nprivate const char * const AccentTab[] = {" >out
	for (i = 0;  i < Naccents;  i++)
		printf("\t\"%s\",\n", AccentTab[Accent[i]]) >out
	print "\tNULL" >out
	print "};" >out
	print fi8 >out
}

#======================================================================
# $Log: Mkctable.awk,v $
# Revision 1.14  1993/11/05  22:22:41  tom
# comment fix.
#
# Revision 1.13  1993/11/04  23:42:17  tom
# (verbose): new option; some comment fixes.
#
# Revision 1.12  1993/01/26  18:39:07  tom
# add RCS Log.
#
#======================================================================
